/**
  *  \file game/load-ov.cc
  *  \brief Directory Overview
  *
  *  Code to generate the directory overview, i.e. to check what's
  *  available.
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "game/load.h"
#include "game/rst.h"
#include "game/trn.h"
#include "util/nls.h"
#include "util/console.h"
#include "util/except.h"
#include "cpluslib/format.h"
#include "io/io.h"
#include "cpluslib/misc.h"
#include "game/gen.h"

static bool
checkResult(Directory& dir, int pid, TResultGen& rgen)
{
    /* Open file */
    string_t fn = format("player%d.rst", pid);
    Ptr<Stream> s = dir.openFileNT(fn, Stream::C_READ);
    if (!s)
        return false;

    /* Check GEN block for validity */
    try {
        GResultFile rst(*s);
        s->seek(rst.getSectionOffset(GResultFile::sec_Gen));
        if (getStructure(*s, rgen)
            && rgen.player_id == pid
            && rgen.turn > 0
            && GGen(rgen).isValidTimestampChecksum())
        {
            /* ok */
            return true;
        } else {
            /* error */
            console.write(LOG_WARN, format(_("File `%s' exists but is invalid -- ignored"), s->getName()));
            return false;
        }
    }
    catch(std::exception& e) {
        console.write(LOG_WARN, fn + ": " + e.what());
        return false;
    }
}


/** \class GDirectoryOverview
    \brief Overview over a directory

    This class scans a directory for usable game files.

    For each player, we can find a result file (have_RST) or an
    unpacked game directory (have_Gamedir), or both belonging to the
    same turn. In addition, there can be a matching TRN (have_TRN).

    In addition, we might find a new RST (have_NewRST).

    If there is a conflict, the conflicting data is marked
    (have_Conflict); essentially, this says one can safely load all
    non-conflicting data. */

/** Construct empty overview. Call scan() to fill it. */
GDirectoryOverview::GDirectoryOverview()
    : global_flags(0)
{
    clear();
}

/** Scan for files. This will populate this object with information
    about the specified directory.

    The basic idea is to scan for unpacked data first, and if none
    found, accept RSTs. PCC2 does not need data be unpacked before
    playing, but if there is unpacked data, we should of course use
    that instead of the RST and a probably older TRN. */
void
GDirectoryOverview::scan(Directory& dir, bool rstOnly)
{
    string_t fn;
    Ptr<Stream> s;
    GTimestamp times[NUM_PLAYERS];
    int  turns[NUM_PLAYERS];
    int  count = 0;

    this->global_flags = 0;
    
    /* do we have unpacked game data? */
    if (!rstOnly) {
        for (int i = 1; i <= NUM_PLAYERS; ++i) {
            turns[i-1] = 0;
            pp_flags[i-1] = 0;
            try {
                fn = format("gen%d.dat", i);
                s = dir.openFileNT(fn, Stream::C_READ);
                if (s) {
                    TGen gen;
                    if (getStructure(*s, gen) && gen.player_id == i && gen.turn > 0) {
                        turns[i-1] = gen.turn;
                        pp_flags[i-1] |= have_Gamedir;
                        times[i-1] = gen.timestamp;
                        ++count;

                        TResultGen rgen;
                        if (checkResult(dir, i, rgen)) {
                            if (rgen.turn == gen.turn && GTimestamp(rgen.timestamp) == GTimestamp(gen.timestamp)) {
                                // same turn, ok
                                pp_flags[i-1] |= have_RST;
                            } else if (rgen.turn > gen.turn && GTimestamp(gen.timestamp).isEarlierThan(GTimestamp(rgen.timestamp))) {
                                // newer turn, ok. PCC 1.x only checks the timestamp.
                                pp_flags[i-1] |= have_NewRST;
                            } else {
                                // we cannot use this RST, so ignore it
                                console.write(LOG_WARN, format(_("File `%s' exists but is too old -- ignored"), s->getName()));
                            }
                        }
                    } else {
                        console.write(LOG_WARN, format(_("File `%s' exists but is invalid -- ignored"), s->getName()));
                    }
                }
            }
            catch(std::exception& e) {
                console.write(LOG_WARN, fn + ": " + e.what());
            }
        }
    }

    if (!count) {
        /* we don't have an unpacked game directory. Do we have a RST? */
        for (int i = 1; i <= NUM_PLAYERS; ++i) {
            TResultGen gen;
            if (checkResult(dir, i, gen)) {
                turns[i-1] = gen.turn;
                times[i-1] = gen.timestamp;
                pp_flags[i-1] |= have_RST;
                ++count;
            }
        }
    }

    if (!count)
        /* directory is empty */
        return;

    /* We have some data. Figure out which is current */
    int current_turn = 0, current_index = 0;
    for (int i = 0; i < NUM_PLAYERS; ++i) {
        if (turns[i] > current_turn) {
            current_turn = turns[i];
            current_index = i;
        }
    }
    ASSERT(current_turn);

    /* Check for conflicts */
    for (int i = 0; i < NUM_PLAYERS; ++i)
        if (turns[i] && (turns[i] != current_turn || times[i] != times[current_index]))
            pp_flags[i] |= have_Conflict;

    /* Last step: look for TRN files. */
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        try {
            fn = format("player%d.trn", i);
            s = dir.openFileNT(fn, Stream::C_READ);
            if (s) {
                GTurnfile tf(*s, false);
                const TTurnHeader& trn = tf.getTurnHeader();
                if (trn.player == i) {
                    /* when the turn is empty or stale, pretend we have none at all */
                    if (trn.num_commands && times[i-1] == trn.timestamp) {
                        pp_flags[i-1] |= have_TRN;
                    }
                } else {
                    console.write(LOG_WARN, format(_("File `%s' exists but is invalid -- ignored"), s->getName()));
                }
            }
        }
        catch(std::exception& e) {
            console.write(LOG_WARN, fn + ": " + e.what());
        }
    }

    /* Now figure out global_flags. */
    for (int i = 0; i < NUM_PLAYERS; ++i)
        global_flags |= pp_flags[i];
}

/** Clear stored state. */
void
GDirectoryOverview::clear()
{
    global_flags = 0;
    std::fill_n(pp_flags, countof(pp_flags), 0);
}

/** Get players that have a particular flag set.
    \param flags Flags to check.
    \return set of all players that have at least one of the requested flags set */
GPlayerSet
GDirectoryOverview::getPlayersWhere(int flags) const
{
    GPlayerSet result;
    for (int i = 1; i <= NUM_PLAYERS; ++i)
        if (pp_flags[i-1] & flags)
            result += i;
    return result;
}

/** Get default player. If this directory contains only one player, return that. */
int
GDirectoryOverview::getDefaultPlayer() const
{
    int result = 0;
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (pp_flags[i-1] != 0) {
            if (result == 0)
                result = i;
            else
                return 0;
        }
    }
    return result;
}
