/**
  *  \file game/maketurn.cc
  *  \brief Maketurn Logic
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <memory>
#include "game/maketurn.h"
#include "cpluslib/format.h"
#include "game/gen.h"
#include "game/load.h"
#include "game/msg.h"
#include "game/playerset.h"
#include "game/racename.h"
#include "io/io.h"
#include "io/textfile.h"
#include "util/console.h"
#include "util/except.h"
#include "util/misc.h"
#include "util/nls.h"

namespace {
    struct ObjInfo {
        int id;
        char fcode[3];
    };
}

/** Generate turn commands for one data file pair.
    \param dir        [in] directory to work in
    \param basename   [in] basename of file ("ship")
    \param pid        [in] player Id
    \param turn       [out] commands are generated here
    \param make       [in] routine to generate commands from memory buffer
    \param recordSize [in] size of records in this file
    \param idOfs      [in] offset of Id number in each record
    \param info       [out] information of first ship is reported here if non-null

    MISRA probably hates me for this function signature. */
static void
generateTurnCommands(Directory& dir,
                     const char basename[], const int pid,
                     GTurnfile& turn,
                     void (GTurnfile::*make)(int,const char*,const char*),
                     const int recordSize, const int idOfs,
                     ObjInfo* info)
{
    /* Allocate buffer */
    static const int N = 156;
    char datBuffer[N], disBuffer[N];
    CTASSERT(N >= TShip::size);
    CTASSERT(N >= TStarbase::size);
    CTASSERT(N >= TPlanet::size);

    /* Open files */
    Ptr<Stream> datFile = dir.openFile(format("%s%d.dat", basename, pid), Stream::C_READ);
    Ptr<Stream> disFile = dir.openFile(format("%s%d.dis", basename, pid), Stream::C_READ);

    /* Read and compare counts */
    TInt16 datCount, disCount;
    getStructureT(*datFile, datCount);
    getStructureT(*disFile, disCount);
    if (datCount.value != disCount.value)
        throw FileFormatException(*datFile, _("Object count mismatch"));

    /* Read and process content */
    for (int i = 0; i < datCount.value; ++i) {
        datFile->readT(datBuffer, recordSize);
        disFile->readT(disBuffer, recordSize);
        int datId = getInt16(&datBuffer[idOfs]);
        int disId = getInt16(&disBuffer[idOfs]);
        if (datId != disId)
            throw FileFormatException(*datFile, _("Object Id mismatch"));
        (turn.*make)(datId, disBuffer, datBuffer);

        /* Remember first object's Id */
        if (info != 0) {
            info->id = datId;
            std::memcpy(info->fcode, &datBuffer[4], 3);
            info = 0;
        }
    }
}

/** Send a single message into a turn file.
    FIXME: copy from save-trn.cc */
static void
sendMessage(GTurnfile& trn, int from, int to, string_t text)
{
    encodeMessage(text);

    char buffer[4];
    storeUint16(buffer + 0, from);
    storeUint16(buffer + 2, to);

    trn.addCommand(tcm_SendMessage, text.length(), buffer, sizeof buffer);
    trn.addData(text.data(), text.length());
}

/** Send a message outbox into a turn file.
    FIXME: copy of from save-trn.cc */
static void
sendOutbox(GTurnfile& trn, GOutbox& outbox)
{
    for (int i = 0; i < outbox.getCount(); ++i) {
        GPlayerSet rec = outbox.getReceiverMask(i);
        for (int p = 1; p <= NUM_PLAYERS+1; ++p) {
            if (rec.contains(p))
                sendMessage(trn, outbox.getSender(), p,
                            outbox.getSendPrefix(i, p) + outbox.getRawText(i));
        }
    }
}

/** Send THost alliance commands. For THost, the mere presence of a
    friendly code change command triggers an alliance action. When we
    change the FCode back to what it should be, we can set alliances
    without sacrificing a ship for each action.
    FIXME: almost copy from save-trn.cc */
static void
sendTHostAllies(GTurnfile& trn, string_t s, const ObjInfo& info)
{
    char buffer[3];

    while (s.length() > 0) {
        storeBasicString(buffer, s);
        trn.addCommand(tcm_ShipChangeFc, info.id, buffer, 3);
        if (s.length() <= 3)
            break;
        s.erase(0, 3);
    }
    trn.addCommand(tcm_ShipChangeFc, info.id, info.fcode, 3);
}


namespace {
    class MessageSender : public GOutboxReader {
        GTurnfile& trn;
     public:
        MessageSender(GTurnfile& trn)
            : trn(trn)
            { }
        void addMessage(string_t text, GPlayerSet receivers)
            {
                // FIXME: this is not very efficient, and it mangles the message through
                // our whole normalisation process. Since we currrently pull in msg-out.o
                // anyway, it doesn't hurt to use GOutbox, but it would still be much
                // better to send messages directly. We'd need some way to deal with
                // recipient headers, though.
                GOutbox out(trn.getPlayer());
                out.addMessage(text, receivers);
                sendOutbox(trn, out);
            }
    };
}

/** Send command messages.
    \param turn      [in/out] Turn to send them on
    \param s         [in] cmdX.txt file
    \param firstShip [in] Information about a ship, for THost allies */
static void
generateCommands(GTurnfile& turn, Stream& s, const ObjInfo& firstShip)
{
    TextFile tf(s);
    string_t line;
    string_t accum;
    while (tf.getLine(line)) {
        line = strTrim(line);
        if (line.size() == 0 || line[0] == '#') {
            // ignore
        } else {
            string_t verb = strNthWord(line, 0);
            if (stringMatch("$TIMEstamp", verb)) {
                // check timestamp
                if (strNthWord(line, 1) != turn.getTimestamp().getTimestampAsString())
                    break;
            } else if (strCaseCmp("$THOST-ALLIES", verb) == 0) {
                // sent THost alliances
                if (firstShip.id == 0) {
                    console.write(LOG_WARN, format(_("Player %d has no ship; alliance changes not transmitted"),
                                                   turn.getPlayer()));
                } else {
                    sendTHostAllies(turn, strNthWordRest(line, 1), firstShip);
                }
            } else if (verb.size() == 0 || verb[0] == '$') {
                // meta-verb or invalid
            } else {
                // command
                if (accum.size() != 0 && accum.size() + line.size() > 500) {
                    sendMessage(turn, turn.getPlayer(), turn.getPlayer(), accum);
                    accum.clear();
                }
                accum += line;
                accum += '\n';
            }
        }
    }

    if (accum.size() != 0)
        sendMessage(turn, turn.getPlayer(), turn.getPlayer(), accum);
}

/** Generate almost complete turn. The turn will be complete except for the
    checksum block; we need all turns to fill in that.
    \param dir Directory
    \param pid Player id
    \return newly-allocated GTurnfile object */
static GTurnfile*
makeProtoTurn(Directory& dir, int pid)
{
    /* Read gen.dat file */
    Ptr<Stream> s = dir.openFile(format("gen%d.dat", pid), Stream::C_READ);
    TGen rawGen;
    getStructureT(*s, rawGen);
    GGen parsedGen(rawGen);
    s = 0;

    /* Create turn */
    std::auto_ptr<GTurnfile> turn(new GTurnfile(pid, parsedGen.getTimestamp()));
    turn->setFeatures(GTurnfile::trnf_Winplan);
    turn->setRegInfo(::getRegInfo(), parsedGen.getTurnNumber());

    /* Generate bulk turn commands */
    ObjInfo firstShip = { 0, "" };
    generateTurnCommands(dir, "ship",   pid, *turn, &GTurnfile::makeShipCommands,   TShip::size,     0, &firstShip);
    generateTurnCommands(dir, "pdata",  pid, *turn, &GTurnfile::makePlanetCommands, TPlanet::size,   2, 0);
    generateTurnCommands(dir, "bdata",  pid, *turn, &GTurnfile::makeBaseCommands,   TStarbase::size, 0, 0);

    /* Messages */
    s = dir.openFileNT(format("mess35%d.dat", pid), Stream::C_READ);
    if (s) {
        MessageSender(*turn).loadOutbox35(*s);
    } else {
        s = dir.openFileNT(format("mess%d.dat", pid), Stream::C_READ);
        if (s) {
            MessageSender(*turn).loadOutbox(*s);
        }
    }

    /* Commands */
    s = dir.openFileNT(format("cmd%d.txt", pid), Stream::C_READ);
    if (s)
        generateCommands(*turn, *s, firstShip);

    /* Password */
    if (rawGen.new_password_flag == 13)
        turn->addCommand(tcm_ChangePassword, 0, rawGen.new_password, sizeof(rawGen.new_password));

    turn->update();

    return turn.release();
}

/******************************* GMaketurn *******************************/

/** Constructor. Makes a blank object.
    \param dir Directory to work in */
GMaketurn::GMaketurn(Directory& dir)
    : dir(dir)
{ }

/** Destructor. */
GMaketurn::~GMaketurn()
{ }

/** Generate turn for a player. The turn is generated in memory; this function
    does not write anything to the disk. Call this exactly once for every player
    in the game directory.
    \param pid Player number [1,NUM_PLAYERS]
    \return Number of commands in that turn */
int
GMaketurn::makeTurn(int pid)
{
    GTurnfile* tf = makeProtoTurn(dir, pid);
    turns.push_back(tf);
    return tf->getNumCommands();
}

/** Finish and write out turn files. */
void
GMaketurn::writeTurns()
{
    /* Fill in all headers */
    uint32_t cksums[NUM_PLAYERS] = { };
    for (unsigned i = 0; i < turns.size(); ++i) {
        cksums[turns[i]->getPlayer()-1] = turns[i]->getDosTrailer().checksum;
    }
    for (unsigned i = 0; i < turns.size(); ++i) {
        turns[i]->setTemplock(cksums);
        turns[i]->updateTrailer();
    }

    /* Write them out */
    for (unsigned i = 0; i < turns.size(); ++i) {
        Ptr<Stream> s = dir.openFile(format("player%d.trn", turns[i]->getPlayer()), Stream::C_WRITE | Stream::C_CREATE);
        turns[i]->write(*s);
    }
}
