/**
  *  \file gfx/gfxaa.cc
  *  \brief Anti-Aliased Graphics Primitives
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "gfx/gfxaa.h"
#include "gfx/gfx.h"
#include "cpluslib/types.h"
#include "gfx/gfxcontext.h"

/** Line drawing, with anti-aliasing. Draws a line from x1,y1 to
    x2,y2, inclusive. For simplicity, this maps all lines into four
    octants, and decides between two cases handling two of them each.
    To allow for lines that nicely connect to each other, we'd need to
    exclude x2,y2 from drawing, and implement eight octants (as four
    cases). The initialisation of the error term might also need
    adjustment.

    \todo does not correctly support alpha != GFX_ALPHA_OPAQUE */
void
drawLineAA(GfxContext& ctx, const GfxPoint p1, const GfxPoint p2)
{
    int x1 = p1.x, y1 = p1.y;
    int x2 = p2.x, y2 = p2.y;

    /* Handle easy cases first */
    if (x1 == x2) {
        drawVLine(ctx, x1, y1, y2);
        return;
    }
    if (y1 == y2) {
        drawHLine(ctx, x1, y1, x2);
        return;
    }

    GfxCanvas& can = ctx.getCanvas();
    uint32_t color = ctx.getRawColor();

    /* Vanilla Bresenham */
    int dx = x2 - x1;
    int dy = y2 - y1;
    if(dx < 0)
        dx = -dx;
    if(dy < 0)
        dy = -dy;

    if(dx < dy) {                // Y is the major axis
        if(y1 > y2) {
            std::swap(y1, y2);
            std::swap(x1, x2);
        }

        int error = dy >> 1;
        int addx  = (x2 < x1) ? -1 : 1;

        /* The error term is the distance between the actual exact
           line and the pixel we set. It directly maps to an alpha
           value. Since our alpha runs from 0..255, the mapping is
           alpha=255*error/dy. To avoid dividing in a loop, we
           represent alpha as its integer part, and the fractional
           part as numerator of a fraction with base dy, and increment
           that step by step. */

        // Initial 'alpha = 255 * error / dy'. Can probably be simplified.
        uint8       alpha       = 255 * error / dy;
        int         alpha_fract = 255 * error % dy;

        // alpha increments by 255*dx/dy each iteration.
        uint8       alpha_inc       = 255 * dx / dy;
        int         alpha_fract_inc = 255 * dx % dy;

        while(y1 <= y2) {
            // draw pixel
            can.drawPixel(GfxPoint(x1,      y1), color, ~alpha);
            can.drawPixel(GfxPoint(x1+addx, y1), color,  alpha);

            // increment y, and error term (fractional X position)
            y1++;
            error += dx;

            // increment alpha
            alpha += alpha_inc;
            alpha_fract += alpha_fract_inc;
            if (alpha_fract >= dy) {
                alpha_fract -= dy;
                ++alpha;
            }

            if(error >= dy) {
                /* decrementing error by dy means decrementing alpha
                   by 255*dy/dy, i.e. incrementing by one */
                error -= dy;
                ++alpha;
                x1 += addx;
            }
        }
    } else {                // X is the major axis
        if(x1 > x2) {
            std::swap(y1, y2);
            std::swap(x1, x2);
        }

        int error = dx >> 1;
        int addy  = (y2 < y1) ? -1 : 1;

        uint8       alpha       = 255 * error / dx;
        int         alpha_fract = 255 * error % dx;

        uint8       alpha_inc       = 255 * dy / dx;
        int         alpha_fract_inc = 255 * dy % dx;

        while(x1 <= x2) {
            can.drawPixel(GfxPoint(x1, y1),      color, ~alpha);
            can.drawPixel(GfxPoint(x1, y1+addy), color,  alpha);
            x1++;
            error += dy;
            alpha += alpha_inc;
            alpha_fract += alpha_fract_inc;
            if (alpha_fract >= dx) {
                alpha_fract -= dx;
                ++alpha;
            }
            if(error >= dx) {
                error -= dx;
                ++alpha;
                y1 += addy;
            }
        }
    }
}

/** Circle drawing, with anti-aliasing. Draws a circle of radius r
    around x0,y0. We draw pixels closer than r+1 from center.

    \todo does not support alpha != GFX_ALPHA_OPAQUE */
void
drawCircleAA(GfxContext& ctx, const GfxPoint pt, int r)
{
    const int x0 = pt.x;
    const int y0 = pt.y;
    register int w = 0;
    register int k = 1;
    register int z = 0;
    register int y2 = 0;
    register int xk = 2*r-1;
    register int x = r;

    /* Given x, we need to compute y = sqrt(r**2 - x**2), or
       y**2 = r**2 - x**2. Starting with x=r, y starts at 0.

       Define z = w**2, and z >= y**2.

       We can incrementally compute exact values for y**2 = y2 using
       the identity that y**2 = \sum_{i=0..y} 2y-1 (bottom of outer loop).

       To find the smallest z such that z >= y**2, we compute the z and w
       values using the same identity (inner loop).

       This generates an y value for all x \in [0,r]. Since we're only
       interested in one octant, we filter out cases where w<x. */
    GfxCanvas& can = ctx.getCanvas();
    uint32_t color = ctx.getRawColor();
    while (x >= 0) {
        while (z < y2) {
            z += k;
            ++w;
            k += 2;
        }
        if (w >= x) {
            /* The actual circle y2 lies between z and next z, i.e.
               z and z+k. Look where we are. This is not mathematically
               clean (should have some roots and squares here), but
               looks well in practice.

               FIXME: should get rid of this division instruction
               someday */
            uint8 alpha = (z - y2) * 255 / k;

            can.drawPixel(GfxPoint(x0 - x, y0 - w), color, ~alpha);
            can.drawPixel(GfxPoint(x0 - x, y0 - w+1), color, alpha);

            can.drawPixel(GfxPoint(x0 + x, y0 - w), color, ~alpha);
            can.drawPixel(GfxPoint(x0 + x, y0 - w+1), color, alpha);

            can.drawPixel(GfxPoint(x0 - w, y0 - x), color, ~alpha);
            can.drawPixel(GfxPoint(x0 - w+1, y0 - x), color, alpha);

            can.drawPixel(GfxPoint(x0 + w, y0 - x), color, ~alpha);
            can.drawPixel(GfxPoint(x0 + w-1, y0 - x), color, alpha);

            can.drawPixel(GfxPoint(x0 - w, y0 + x), color, ~alpha);
            can.drawPixel(GfxPoint(x0 - w+1, y0 + x), color, alpha);

            can.drawPixel(GfxPoint(x0 + w, y0 + x), color, ~alpha);
            can.drawPixel(GfxPoint(x0 + w-1, y0 + x), color, alpha);

            can.drawPixel(GfxPoint(x0 - x, y0 + w), color, ~alpha);
            can.drawPixel(GfxPoint(x0 - x, y0 + w-1), color, alpha);

            can.drawPixel(GfxPoint(x0 + x, y0 + w), color, ~alpha);
            can.drawPixel(GfxPoint(x0 + x, y0 + w-1), color, alpha);
        }
        y2 += xk;
        xk -= 2;
        --x;
    }
}
