/**
  *  \file int/if/ifutil.cc
  *  \brief Interpreter Game Interface: Utilities
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cctype>
#include <cmath>
#include "int/if/ifutil.h"
#include "int/interror.h"
#include "int/value.h"
#include "game/cspec.h"

/** Look up a name in a mapping list.
    \param name    Name of property to find
    \param mapping Mapping list, must be sorted
    \param size    Size of mapping list
    \return index such that mapping[return].name == name, or -1 */
int32_t
lookupName(const string_t name, const TIntMapping* mapping, const uint32_t size)
{
    uint32_t low = 0;
    uint32_t high = size;

    // Binary intersection
    while (high - low > 3) {
        uint32_t mid = (high + low) / 2;
        if (name < mapping[mid].name)
            high = mid;
        else
            low = mid;
    }

    // Linear search
    while (low < high) {
        if (name == mapping[low].name) {
            return low;
        }
        ++low;
    }
    return -1;
}

/** Check integer argument. Note that this also accepts float arguments,
    like PCC 1.x's EvalArgs().
    \param out   [out] Result will be placed here
    \param value [in] Value given by user
    \return true if value was specified, false if value was null (out not changed)
    \throw IntError if value is invalid */
bool
checkIntArg(int32_t& out, IntValue* value)
{
    if (value == 0)
        return false;

    if (IntIntValue* iv = dynamic_cast<IntIntValue*>(value)) {
        // Regular integer.
        out = iv->getValue();
        return true;
    } else if (IntFloatValue* fv = dynamic_cast<IntFloatValue*>(value)) {
        // We truncate here. This is what PCC 1.x does. Let's hope that IEEE FP is precise
        // enough that we don't lose enough precision to get off-by-one results, like HOST.EXE.
        double v = fv->getValue();
        if (std::fabs(v) > 2147483647.0)
            throw IntError::rangeError();
        out = int32_t(v);
        return true;
    } else {
        throw IntError::typeError(IntError::tyInteger);
    }
}

/** Check integer argument with range.
    \param out   [out] Result will be placed here
    \param value [in] Value given by user
    \param min   [in] Minimum accepted value
    \param min   [in] Maximum accepted value
    \return true if value was specified, false if value was null (out not changed)
    \throw IntError if value is invalid */
bool
checkIntArg(int32_t& out, IntValue* value, int32_t min, int32_t max)
{
    bool result = checkIntArg(out, value);
    if (result) {
        if (out < min || out > max)
            throw IntError::rangeError();
    }
    return result;
}

/** Check bool argument.
    \param out   [out] Result will be placed here
    \param value [in] Value given by user
    \return true if value was specified, false if value was null (out not changed)
    \throw IntError if value is invalid (can currently not happen) */
bool
checkBoolArg(bool& out, IntValue* value)
{
    int n = getBoolValue(value);
    if (n < 0)
        return false;
    out = (n != 0);
    return true;
}

/** Check string argument.
    \param out   [out] Result will be placed here
    \param value [in] Value given by user
    \return true if value was specified, false if value was null (out not changed)
    \throw IntError if value is invalid (can currently not happen) */
bool
checkStringArg(string_t& out, IntValue* value)
{
    if (value == 0)
        return false;

    out = value->toString(false);
    return true;
}

/** Check flag argument.
    Users specify flags as a string containing latin letters. Optionally, if
    valueOut is given, an additional integer argument can be specified.
    The set of possible flags is specified using a list of upper-case letters.
    The result will have the "1<<n" bit set if flag "tpl[n]" was specified.
    \param flagOut  [out] Flag result will be placed here
    \param valueOut [out,optional] Value result will be placed here
    \param value    [in] Value given by user
    \param tpl      [in] Template used to parse user's input
    \return true if value was specified, false if value was null (flagOut, valueOut not changed)
    \throw IntError if value is invalid */
bool
checkFlagArg(int32_t& flagOut, int32_t* valueOut, IntValue* value, const char* tpl)
{
    if (value == 0)
        return false;

    if (IntIntValue* iv = dynamic_cast<IntIntValue*>(value)) {
        // An integer fills the valueOut, and specifies no flags
        if (!valueOut)
            throw IntError::typeError();
        *valueOut = iv->getValue();
        flagOut = 0;
        return true;
    } else if (IntStringValue* sv = dynamic_cast<IntStringValue*>(value)) {
        // Parse string
        using std::toupper;
        const string_t& s = sv->getValue();
        uint32_t parsedValue = 0;
        enum { No, Yes, Followed } valueState = No;
        for (string_t::size_type i = 0; i < s.size(); ++i) {
            char c = toupper(uint8_t(s[i]));
            if (const char* p = strchr(tpl, c)) {
                // Flag
                flagOut |= 1 << (p - tpl);
                if (valueState == Yes)
                    valueState = Followed;
            } else if (c >= '0' && c <= '9') {
                // Number
                if (valueState == Followed || !valueOut)
                    throw IntError("Invalid option");
                parsedValue = 10*parsedValue + c - '0';
                *valueOut = parsedValue;
                valueState = Yes;
            } else {
                // Error
                throw IntError("Invalid option");
            }
        }
        return true;
    } else {
        throw IntError::typeError();
    }
}

/** Check cargospec argument.
    \param out   [out] Result will be placed here
    \param value [in] Value given by user
    \return true if value was specified, false if value was null (out not changed)
    \throw IntError if value is invalid (can currently not happen) */
bool
checkCargoSpecArg(GCargoSpec& out, IntValue* value)
{
    if (value == 0)
        return false;

    out = GCargoSpec(value->toString(false));
    return true;
}
