/**
  *  \file resmgr/res-wpvcr.cc
  *  \brief Winplan WPVCR.DLL Resource Provider
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "resmgr/res-wpvcr.h"
#include "util/nls.h"
#include "util/except.h"
#include "io/io.h"
#include "cpluslib/assert.h"
#include "cpluslib/format.h"
#include "resmgr/resid.h"

/** \class ResProviderWinplanVcr
    \brief Winplan WPVCR.DLL Resource Provider

    This class provides functionality for reading the WPVCR.DLL file
    from Winplan. That file contains pictures to use in the VCR
    (left/right moving small ships). */

ResProviderWinplanVcr::ResProviderWinplanVcr(Ptr<Stream> file)
    // We need to keep a copy of 'file' for lifetime management
    : worker(file),
      file(new SubStream(*worker))
{
    init();
}

ResProviderWinplanVcr::ResProviderWinplanVcr(string_t name)
    : worker(Stream::openFile(name, Stream::C_READ)),
      file(new SubStream(*worker))
{
    init();
}

ResProviderWinplanVcr::~ResProviderWinplanVcr()
{ }

/** Initializer. Load the directories and configure the SubStream
    object. */
void
ResProviderWinplanVcr::init()
{
    ASSERT(file);
    char buffer[ENTRIES * 4];
    file->setDisallowedOperations(Stream::C_WRITE);
    file->setRange(0, -1);

    for (int i = 0; i < 2; ++i) {
        file->seek(1000 * i);
        if (file->read(buffer, sizeof buffer) != sizeof buffer)
            throw FileFormatException(*file, _("File is invalid"));
        getUint32(buffer, offsets[i], ENTRIES);
        
        file->seek(1000 * i + 649);
        if (file->read(buffer, sizeof buffer / 2) != sizeof buffer / 2)
            throw FileFormatException(*file, _("File is invalid"));
        getUint16(buffer, sizes[i], ENTRIES);

        for (int j = 0; j < ENTRIES; ++j)
            if (offsets[i][j] && sizes[i][j])
                --offsets[i][j];
    }
}

Ptr<GfxPixmap>
ResProviderWinplanVcr::loadPixmap(ResId id)
{
    int side;

    if (id.isElem(res::rship, 1))
        side = 0;
    else if (id.isElem(res::lship, 1))
        side = 1;
    else
        return 0;

    /* we have 160 ship images only */
    if (id.getId1() <= 0 || id.getId1() > ENTRIES)
        return 0;

    /* do we have that one? */
    int index = id.getId1() - 1;
    if (sizes[side][index] && offsets[side][index]) {
        file->setRange(offsets[side][index], sizes[side][index]);
        Ptr<GfxPixmap> pix = loadImage(file);
        if (pix) {
            /* I was cropping the image here, but I think that is
               counter-productive. It messes up relations between
               images, which would be nice to preserve for FLAK. */
            // FIXME: can we assume this color-keying? I hope, yes. It fails if someone
            // makes a Cube that fills out its complete bitmap.
            pix->setColorKey(pix->getPixel(GfxPoint(0, 0)));
            return pix;
        }
    }
    return 0;
}
