#!/usr/bin/perl -w
#
#  PCC2 Assembler
#
#  What's it good for? Mostly testing.
#
#  Invocation:
#    c2asm.pl inputfile.qs [outputfile.qc]
#  This outputs a PCC2 object file which currently only 'teststmt' can read.
#
#  Syntax of input: sequence of subroutine definitions
#    sub foo(args)
#       assembly-insn
#     label:
#       assembly-insn
#       assembly-insn
#    endsub
#  The header has the same form as in regular CCScript; in particular, the
#  'Optional' keyword can be used, and 'Function' instead of 'Sub' says
#  this routine returns a value (the header line actually just preinitializes
#  the local table ('.local'), and the '.flags', '.min_args', '.max_args').
#
#  Comments can appear anywhere and start with ';' or '%'. Everything is
#  case-insensitive.
#
#  Assembly insns have their regular name and take arguments in their
#  regular form:
#  - integers (e.g. 'pushint', 'dup')
#  - symbol name (e.g. 'pushvar'). The symbol is added to the symbol pool.
#  - local name (e,g. 'pushloc'). The argument is a local name which must
#    exist.
#  - subroutine name (e.g. 'pushsub'). See below.
#  - label name (e.g. 'j'). The label must be defined somewhere in this sub.
#  - literals (argument for 'pushlit'):
#    . integers, floats
#    . 'true', 'false', 'null' (note that these normally are encoded using
#      'pushbool', not 'pushlit')
#    . strings in single or double quotes; backslash escapes in
#      double-quoted strings
#    . '(tag,value)' generates a raw tag/value node for interesting effects,
#      e.g. '(130,5)' generates 'Minefield(5)'.
#    . By default, duplicate literals are recycled. Prefix with '!' or 'new'
#      to create a new instance.
#  - as a special exception, '#nn' is always accepted and generates the
#    argument specified using the integer 'nn', even if the insn doesn't
#    expect an argument, i.e. 'pushlit #5' pushes the 5th literal; 'uinc #5'
#    generates a "inc" unary instruction with 5 in the (unused) arg field.
#  To generate arbitrary opcodes, use 'genTYPMAJOR.MINOR', where TYP is the
#  kind of the argument expected ('int', 'loc', 'sym', 'sub', 'label', 'lit')
#  and MAJOR/MINOR are the decimal opcode. If no TYP is given, the instruction
#  takes an optional 'int' argument. Also, the assembler accepts some invalid
#  instructions that have a valid encoding, e.g. 'procmem', 'popint' or
#  'dimlit'.
#
#  Pseudo-ops:
#  - in subroutines
#    .local NAME      add local variable ('.local -' to make a nameless one)
#    .lit LITERAL     add literal to literal pool (subject to recycling rules)
#    .sym SYMBOL      add symbol to symbol pool (recycles existing)
#    .sub SUB         add subroutine to sub pool (recycles existing)
#    .name NAME       set this sub's name hint ('.name -' to make it empty)
#    .min_args NN     set this sub's minimum arg count
#    .max_args NN     set this sub's maximum arg count
#    .flags NN        set this sub's flags
#    .defsubs         generate pushsub/sdefsub instructions for all uniquely-
#                     named subs so far, excluding this one
#  - global
#    .jumps abs|sym   set jump mode to absolute (default) or symbolic.
#                     Symbolic mode generates actual 'label' instructions
#                     which gives slower execution but permits the optimizer
#                     to run; absolute mode resolves all jumps.
#
#  Sub naming: each sub has three names:
#  - assembler name, which is given in 'sub foo'
#  - name hint, which defaults to the assembler name and can be changed
#    with '.name'
#  - real name, which is given by the 'sdefsub' instruction that defines the
#    sub.
#  An instruction taking a subroutine name always uses the assembler name.
#  In case of a duplicate, the name refers to the most-recently defined one.
#  To define multiple subs with the same name, and still be able to refer
#  to both from the assembler, use the ".name" instruction. This CCScript
#    If a Then
#      Sub foo
#        ...
#      EndSub
#    Else
#      Sub foo
#        ...
#      EndSub
#    EndIf
#  can be assembled to
#    sub foo_one
#      .name foo
#      ...
#    endsub
#    sub foo_two
#      .name foo
#      ...
#    endsub
#    sub main
#      pushvar a
#      jfep else
#      pushsub foo_one
#      j endif
#     else:
#      pushsub foo_two
#     endif:
#      sdefsub foo
#    endsub
#  Note that the '.name' is actually optional as it only affects optimisation
#  in the interpreter (code of identically named subs will be merged if it is
#  identical to avoid duplication across save/reload cycles). The script will
#  work correctly even without name hints.
#
#  The last subroutine defined in the assembler file will be invoked when
#  loading the object file. No subroutine will be automatically defined in the
#  symbol table. Therefore, the last routine must enter them into the symbol
#  table using 'pushsub/sdefsub' instruction pairs. To simplify this, you can
#  use the '.defsubs' pseudo-op, which will generate all appropriate
#  definitions.
#
#  Hacked together in July 2010 for PCC2 beta 10 by Stefan Reuther.
#
#  TODO:
#  - generate debug information (i.e. '.line' and '.file' pseudo-ops)
#  - limit checks: no more than 64k literals, names, labels/instructions,
#    subroutines or locals per subroutine (those checks are currently missing
#    in PCC2 as well); no symbols over 255 characters
#

use strict;
use Data::Dumper; # for debugging

# Description of the instructions
#   mnemonic -> [major, minor, argPI]
# - major and minor are the opcodes
# - minor is the arg parsing information
#   . none (no arg, use 0)
#   . int (16-bit integer)
#   . sym (symbol from name list)
#   . loc (local variable from local name list)
#   . sub (subroutine referred to from sub list)
#   . lit (literal from literal pool)
#   . label (label within this routine)
my %instructionTable;

# All compiled subroutines. Indexed by name, contains references to hashes:
# - id                  (=object Id)
# - name                (=name hint, and used to reference it)
# - flags               (=flags for header property)
# - min_args, max_args  (=argument ranges)
# - num_labels          (=number of symbolic labels)
# - datahead, databody  (=list of literals)
# - names               (=list of names)
# - code                (=list of integers containing the code)
# - subrs               (=list of ids of referenced subrs)
# - locals              (=list of local names)
my @compiledSubs;

# Subroutines by name. This is used to resolve 'sub' operands.
my %subsByName;

# Counter for subroutines, used to generate Ids.
my $subIdCounter = 0;

# Jump mode. Nonzero for symbolic jumps, zero for absolute.
my $symbolicJumps = 0;

# Parse arguments
my $outFN;
my $inFN;
foreach (@ARGV) {
    if (/^-./) {
        die "$0: Unknown option '$_'\n"
    } elsif (!defined($inFN)) {
        $inFN = $_;
    } elsif (!defined($outFN)) {
        $outFN = $_;
    } else {
        die "$0: Too many arguments\n";
    }
}
if (!defined($inFN)) {
    die "$0: Need an input file name\n"
}
if (!defined($outFN)) {
    $outFN = $inFN;
    $outFN =~ s/\.[^.]+$//;
    $outFN .= ".qc";
}
if ($inFN eq $outFN) { die "$0: Input file name is same as output file name\n" }

# Initialize
initInstructionTable();

# Assembler main loop.
open IN, "< $inFN" or die "$inFN: $!\n";
print "Assembling $inFN...\n";
my $numErrors = 0;
while (<IN>) {
    preprocessLine();
    if (/^(sub|function)\s+([0-9a-zA-Z.\$_]+)\s*(.*)/i) {
        # subroutine definition
        my $kind = uc($1);
        my $name = uc($2);
        my $args = $3;
        my $theSubroutine = { id => ++$subIdCounter,
                              name => $name,
                              flags => ($kind eq 'SUB' ? 1 : 0),
                              min_args => 0,
                              max_args => 0,
                              num_labels => 0,
                              datahead => [],
                              databody => [],
                              names => [],
                              code => [],
                              subrs => [],
                              locals => [] };

        # does it have args?
        if ($args =~ /^\((.*)\)$/) {
            my $opt = 0;
            foreach (split /,\s*/, $1) {
                my $name = uc($_);
                if (/^optional\s+(.*)/) {
                    $name = uc($1);
                    $opt = 1;
                }
                push @{$theSubroutine->{locals}}, $name;
                ++$theSubroutine->{max_args};
                if (!$opt) { ++$theSubroutine->{min_args} };
            }
        } elsif ($args ne '') {
            err("invalid argument specification");
        }

        # log it
        # if ($theSubroutine->{min_args} != $theSubroutine->{max_args}) {
        #     printf "  %4d   %s (%d..%d args)\n",
        #       $theSubroutine->{id},
        #         $theSubroutine->{name},
        #           $theSubroutine->{min_args},
        #             $theSubroutine->{max_args};
        # } else {
        #     printf "  %4d   %s (%d args)\n",
        #       $theSubroutine->{id},
        #         $theSubroutine->{name},
        #           $theSubroutine->{min_args};
        # }

        # Assemble it
        assemble($theSubroutine);

        # Remember it
        push @compiledSubs, $theSubroutine;
        $subsByName{$name} = $theSubroutine;
    } elsif (/^\.jumps\s+(\S..+)/i) {
        # configure jump mode for following subs
        my $mode = lc($1);
        if (substr("absolute", 0, length($mode)) eq $mode) {
            $symbolicJumps = 0;
        } elsif (substr("symbolic", 0, length($mode)) eq $mode) {
            $symbolicJumps = 1;
        } else {
            err("invalid jump mode");
        }
    } elsif (/^$/) {
        # ignore empty line
    } else {
        err("invalid directive");
    }
}
close IN;

if ($numErrors > 0) {
    die "Input has errors; output file not created.\n";
}

# The entry point always is the last sub. Anything else would be pointless
# because only the last sub can refer to (and thus define) all others.
if (@compiledSubs == 0) {
    die "input does not contain any code";
}
my $entry = $compiledSubs[-1]{id};

# Generate output
print "Writing $outFN (".scalar(@compiledSubs)." subs)...\n";
open OUT, "> $outFN" or die "$outFN: $!\n";
print OUT "CCobj\032";
print OUT pack("vvV", 100, 4, $entry);
foreach (@compiledSubs) {
    print OUT packSub($_);
}
exit 0;

############################### Subroutines ###############################

# Assemble a subroutine
#   $theSubroutine = target
sub assemble {
    my $theSubroutine = shift;
    my @relocations;
    my %labels;
    while (<IN>) {
        preprocessLine();

        # Check for labels
        while (s/^(\S+):\s*//) {
            my $name = uc($1);
            if (exists $labels{$name}) { err("duplicate label '$name'"); next }

            # Generate the label
            my $address;
            if ($symbolicJumps) {
                $address = $theSubroutine->{num_labels}++;
                push @{$theSubroutine->{code}}, 0x04800000 + $address;
            } else {
                $address = scalar(@{$theSubroutine->{code}});
            }
            $labels{$name} = $address;
        }

        # Anything here?
        if (/^\s*$/) { next }

        # Split into mnemonic and arg
        my $mnemo = lc($_);
        my $arg = "";
        if (/^(.*?) (.*)/) {
            $mnemo = lc($1);
            $arg = $2;
        }

        # Check for end
        if ($mnemo eq 'endsub' || $mnemo eq 'endfunction') { last }

        # Check for specials
        if ($mnemo eq '.local') {
            if ($arg eq '') { err("'.local' needs a name ('-' for none)"); next }
            push @{$theSubroutine->{locals}}, ($arg eq '-' ? '' : uc($arg));
            next;
        }
        if ($mnemo eq '.lit') {
            if ($arg eq '') { err("'.lit' needs an argument"); next }
            checkAddLiteral($theSubroutine, $arg);
            next;
        }
        if ($mnemo eq '.sym') {
            if ($arg eq '') { err("'.sym' needs an argument"); next }
            checkAddSymbol($theSubroutine, uc($arg));
            next;
        }
        if ($mnemo eq '.sub') {
            if ($arg eq '') { err("'.sub' needs an argument"); next }
            checkAddSub($theSubroutine, uc($arg));
            next;
        }
        if ($mnemo eq '.name') {
            if ($arg eq '') { err("'.name' needs an argument ('-' for none)"); next }
            $theSubroutine->{name} = ($arg eq '-' ? '' : uc($arg));
            next;
        }
        if ($mnemo =~ /^\.(min_args|max_args|flags|num_labels)$/) {
            my $what = $1;
            if ($arg !~ /^\d+$/) { err("'.$what' needs an integer argument"); next }
            $theSubroutine->{$what} = $arg+0;
            next;
        }
        if ($mnemo eq '.defsubs') {
            if ($arg ne '') { err("'.defsubs' doesn't take an argument"); next }
            foreach my $s (sort keys %subsByName) {
                # Generate 'pushsub, sdefsub' for each of these subs
                push @{$theSubroutine->{code}},
                  makeInstruction($instructionTable{pushsub}[0],
                                  $instructionTable{pushsub}[1],
                                  checkAddSub($theSubroutine, $s));
                push @{$theSubroutine->{code}},
                  makeInstruction($instructionTable{sdefsub}[0],
                                  $instructionTable{sdefsub}[1],
                                  checkAddSymbol($theSubroutine, $subsByName{$s}{name}));
            }
            next;
        }

        # Check instructions
        my $major;     # major opcode
        my $minor;     # minor opcode
        my $argPI;     # argument parsing information (=argument type)
        if ($mnemo =~ /^gen(|int|sym|loc|lit|sub|label)(\d+)\.(\d+)$/) {
            # Escape opcode
            $major = $2;
            $minor = $3;
            $argPI = $1 ne '' ? $1 : $arg ne '' ? 'int' : 'none';;
        } elsif (exists $instructionTable{$mnemo}) {
            # Regular
            $major = $instructionTable{$mnemo}[0];
            $minor = $instructionTable{$mnemo}[1];
            $argPI = $instructionTable{$mnemo}[2];
        } else {
            err("invalid instruction '$mnemo'"); next;
        }

        # Optional arg? If $argPI starts with a '?', and the user didn't
        # give an argument, that's ok.
        if ($argPI =~ s/^\?//) {
            if ($arg eq '') { $argPI = 'none' }
        }

        # Parse arg
        my $argVal;
        if ($arg =~ /^\#([-+]?[0-9]+)$/) {
            # override: user wants integer arg no matter what the instruction wants
            $argVal = $1 & 65535;
        } elsif ($argPI eq 'none') {
            # no arg
            if ($arg ne '') { err("instruction doesn't expect argument"); }
            $argVal = 0;
        } elsif ($argPI eq 'int') {
            # integer
            if ($arg !~ /^[-+]?[0-9]+$/) {
                err("instruction expects integer argument");
                $argVal = 0;
            } else {
                $argVal = $arg & 65535;
            }
        } elsif ($argPI eq 'sym') {
            # symbol
            $argVal = checkAddSymbol($theSubroutine, uc($arg));
        } elsif ($argPI eq 'loc') {
            # local name
            $argVal = checkLocal($theSubroutine, uc($arg));
        } elsif ($argPI eq 'lit') {
            # literal
            $argVal = checkAddLiteral($theSubroutine, $arg);
        } elsif ($argPI eq 'sub') {
            # sub
            $argVal = checkAddSub($theSubroutine, uc($arg));
        } elsif ($argPI eq 'label') {
            # label
            if (exists $labels{uc($arg)}) {
                # backward reference: use directly
                $argVal = $labels{uc($arg)};
            } else {
                # forward reference
                push @relocations, [scalar(@{$theSubroutine->{code}}), uc($arg)];
                $argVal = 0;
            }
        } else {
            die "internal error: invalid arg parsing information '$argPI'";
        }

        if ($argPI eq 'label' && $symbolicJumps) {
            $minor |= 0x80
        }

        $argVal &= 65535;
        # printf "  %02X.%02X.%04X\n", $major, $minor, $argVal;
        push @{$theSubroutine->{code}}, makeInstruction($major, $minor, $argVal);
    }

    if (@relocations) {
        # printf "  %d relocations\n", scalar(@relocations);
        foreach (@relocations) {
            if (!exists $labels{$_->[1]}) { err("unknown label '$_->[1]'"); next }
            $theSubroutine->{code}[$_->[0]] += $labels{$_->[1]};
        }
    }
}


# Preprocess line
sub preprocessLine {
    # Remove line ending
    s/[\r\n]+//g;

    # Remove comments (';' or '%')
    #           'xxxxx' "xxxxxxxxxxxxx"
    s/^(([^%'"]|'[^']*'|"([^"\\]|\\.)*")*)[;%].*/$1/;      #"

    # Remove leading/trailing space
    s/^\s*//;
    s/\s*$//;
}

# Initialize %instructionTable
sub initInstructionTable {
    # Initialize
    %instructionTable = (
                         # 0 - push
                         # 1 - binary
                         # 2 - unary
                         # 3 - ternary
                         # 4 - jump
                         nop     => [4,0,'none'],
                         label   => [4,0,'label'],
                         jt      => [4,1,'label'],
                         jf      => [4,2,'label'],
                         jtf     => [4,3,'label'],
                         je      => [4,4,'label'],
                         jte     => [4,5,'label'],
                         jfe     => [4,6,'label'],
                         j       => [4,7,'label'],
                         jneverp => [4,8,'label'],
                         jtp     => [4,9,'label'],
                         jfp     => [4,10,'label'],
                         jtfp    => [4,11,'label'],
                         jep     => [4,12,'label'],
                         jtep    => [4,13,'label'],
                         jfep    => [4,14,'label'],
                         jp      => [4,15,'label'],
                         catch   => [4,16,'label'],
                         jdz     => [4,17,'label'],

                         # 5 - indirect
                         # 6 - stack
                         dup     => [6,0,'int'],
                         drop    => [6,1,'int'],
                         swap    => [6,2,'int'],

                         # 7 - pop
                         # 8 - store
                         # 9 - memref
                         # 10 - dim
                         # 11 - special
                         suncatch    => [11,0,'none'],
                         sreturn     => [11,1,'int'],
                         swith       => [11,2,'none'],
                         sendwith    => [11,3,'none'],
                         sfirstindex => [11,4,'none'],
                         snextindex  => [11,5,'none'],
                         sendindex   => [11,6,'none'],
                         sevals      => [11,7,'int'],
                         sevalx      => [11,8,'none'],
                         sdefsub     => [11,9,'sym'],
                         sdefshipp   => [11,10,'sym'],
                         sdefplanetp => [11,11,'sym'],
                         sload       => [11,12,'none'],
                         sprint      => [11,13,'none'],
                         saddhook    => [11,14,'?int'], # optional arg; might be used later, but isn't used now
                         srunhook    => [11,15,'?int'], # optional arg; might be used later, but isn't used now
                         sthrow      => [11,16,'none'],
                         sterminate  => [11,17,'none'],
                         ssuspend    => [11,18,'none'],
                        );

    # Major 2 - unary
    my @unaryList = qw( not bool neg pos sin
                        cos tan zap abs exp
                        log bitnot isempty isnum isstr
                        asc chr str sqrt trunc
                        round ltrim rtrim lrtrim length
                        val trace not2 atom atomstr
                        keycreate keylookup inc dec isproc );
    foreach (0 .. $#unaryList) {
        $instructionTable{'u'.lc($unaryList[$_])} = [2,$_,'none'];
    }

    # Major 1 - binary
    my @binaryList = qw( and or xor add sub
                         mul div idiv rem pow
                         concat concempty cmpeq cmpeqNC cmpne
                         cmpneNC cmple cmpleNC cmplt cmpltNC
                         cmpge cmpgeNC cmpgt cmpgtNC min
                         minNC max maxNC firststr firststrNC
                         reststr reststrNC findstr findstrNC bitand
                         bitor bitxor str atan lcut
                         rcut endcut strmult keyaddparent keyfind );
    foreach (0 .. $#binaryList) {
        $instructionTable{'b'.lc($binaryList[$_])} = [1,$_,'none'];
    }

    # Major 3 - ternary
    my @ternaryList = qw( keyadd );
    foreach (0 .. $#ternaryList) {
        $instructionTable{'t'.lc($ternaryList[$_])} = [3,$_,'none'];
    }

    # Major 0,7,8,10 - "scope" operations
    foreach (['var',0,'sym'],  ['loc',1,'loc'],  ['top',2,'int'],
             ['glob',3,'int'], ['gvar',4,'sym'], ['lit',5,'lit'],
             ['int',6,'int'],  ['bool',7,'int'], ['sub',8,'sub'])
      {
          $instructionTable{'push'.$_->[0]}  = [0, $_->[1], $_->[2]];
          $instructionTable{'pop'.$_->[0]}   = [7, $_->[1], $_->[2]];
          $instructionTable{'store'.$_->[0]} = [8, $_->[1], $_->[2]];
          $instructionTable{'dim'.$_->[0]}   = [10, $_->[1], 'sym'];
      }

    # Major 5,9 - "i/m" operations
    foreach (["call"  => 0], ["load"  => 1], ["store"  => 2], ["pop"   => 3],
             ["proc"  => 4], ["pload" => 5], ["pstore" => 6], ["ppop"  => 7],
             ["fcall" => 8], ["func"  => 9], ["fstore" => 10], ["fpop" => 11])
      {
          $instructionTable{$_->[0]."ind"} = [5,$_->[1], 'int'];
          $instructionTable{$_->[0]."mem"} = [9,$_->[1], 'sym'];
      }
}

# Compile 'sym' arg: check for symbol, add it to $theSubroutine's symbol table
sub checkAddSymbol {
    my ($theSubroutine, $sym) = @_;
    foreach (0 .. $#{$theSubroutine->{names}}) {
        if ($theSubroutine->{names}[$_] eq $sym) { return $_ }
    }
    push @{$theSubroutine->{names}}, $sym;
    return $#{$theSubroutine->{names}};
}

# Compile 'local' arg: must be defined
sub checkLocal {
    my ($theSubroutine, $loc) = @_;
    foreach (0 .. $#{$theSubroutine->{locals}}) {
        if ($theSubroutine->{locals}[$_] eq $loc) { return $_ }
    }
    err("unknown local");
    return 0;
}

# Compile 'lit' arg: compile it, and ensure it is in the literal table
sub checkAddLiteral {
    my ($theSubroutine, $lit) = @_;
    my $head;
    my $body;
    my $force = 0;
    if ($lit =~ s/^(!\s*|new\s+)//i) { $force = 1 };
    if (lc($lit) eq 'true') {
        # boolean
        $head = pack("vV", 0x300, 1);
        $body = '';
    } elsif (lc($lit) eq 'false') {
        # boolean
        $head = pack("vV", 0x300, 0);
        $body = '';
    } elsif (lc($lit) eq 'null') {
        # null
        $head = pack("vV", 0x100, 0);
        $body = '';
    } elsif ($lit =~ /^[-+]?[0-9]+$/) {
        # integer
        $head = pack("vV", 0x200, $lit);
        $body = '';
    } elsif ($lit =~ /^[-+]?[0-9]*\.[0-9]*$/ && $lit =~ /[0-9]/) {
        # float (second regexp checks that there is at least one digit in it)
        my $value = $lit;
        my $sign = 0;
        if ($value < 0) {
            $value = -$value;
            $sign = 0x80000000;
        }
        if ($value > 1.0e39) {
            # Large value: pack max-real
            $head = pack("vV", 0xFFFF, 0x7FFFFFFF | $sign);
        } elsif ($value == 0) {
            # Zero
            $head = pack("vV", 0, 0);
        } else {
            # Build value. Almost copied from int/value.cc.
            my ($mant, $exp) = frexp($value);
            $exp -= 1;
            $exp += 129;
            if ($exp <= 0) {
                # underflow or zero
                $head = pack("vV", 0, 0);
            } elsif ($exp > 255) {
                # overflow
                $head = pack("vV", 0xFFFF, 0x7FFFFFFF | $sign);
            } else {
                # regular
                $mant -= 0.5;           # 0.0yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyxxxxxxxx
                $mant *= 4294967296.0;  # yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy.xxxxxxxx
                my $bits1 = int($mant);
                $mant -= $bits1;        # 0.xxxxxxxx
                $mant *= 256.0;
                my $bits2 = int($mant);
                $head = pack("vV", $exp | (256*$bits2), $bits1 | $sign);
            }
        }
        $body = '';
    } elsif ($lit eq "''" || $lit eq '""') {
        # empty string; encode as short string
        $head = pack("vV", 0x400, 0);
        $body = '';
    } elsif ($lit =~ /^'(.*)'$/) {
        # regular string
        $head = pack("vV", 0x600, length($1));
        $body = $1;
    } elsif ($lit =~ /^"(.*)"$/) {
        # regular string with escaped items
        my $s = $1;
        $s =~ s/\\(.)/$1/g;
        $head = pack("vV", 0x600, length($s));
        $body = $s;
    } elsif ($lit =~ /^\(\s*(\d+)\s*,\s*(\d+)\s*\)$/) {
        # override
        $head = pack("vV", 0x100*$1, $2);
        $body = "";
    } else {
        err("unparsed literal '$lit'");
        # generate null
        $head = pack("vV", 0x100, 0);
        $body = '';
    }

    # Recycle literal if possible
    if (!$force) {
        foreach (0 .. $#{$theSubroutine->{datahead}}) {
            if ($head eq $theSubroutine->{datahead}[$_] && $body eq $theSubroutine->{databody}[$_]) { return $_ }
        }
    }

    # Add new literal
    push @{$theSubroutine->{datahead}}, $head;
    push @{$theSubroutine->{databody}}, $body;
    return $#{$theSubroutine->{datahead}};
}

# Check for subroutine
sub checkAddSub {
    my ($theSubroutine, $sub) = @_;

    # Convert subroutine into Id
    if (!exists $subsByName{$sub}) { err("unknown sub '$sub'"); return 0 }
    my $subid = $subsByName{$sub}{id};

    # Do we already have it?
    foreach (0 .. $#{$theSubroutine->{subrs}}) {
        if ($theSubroutine->{subrs}[$_] == $subid) { return $_ }
    }

    push @{$theSubroutine->{subrs}}, $subid;
    return $#{$theSubroutine->{subrs}};
}

# Convert subroutine into serialized form.
# Returns a completely packed-up bytecode object.
sub packSub {
    my $theSubroutine = shift;
    my @properties = ([0,
                       ''],
                      [0,
                       pack('vvvv',
                            $theSubroutine->{flags},
                            $theSubroutine->{min_args},
                            $theSubroutine->{max_args},
                            $theSubroutine->{num_labels})],
                      [scalar(@{$theSubroutine->{datahead}}),
                       join('',
                            @{$theSubroutine->{datahead}},
                            @{$theSubroutine->{databody}})],
                      [scalar(@{$theSubroutine->{names}}),
                       packStringList(@{$theSubroutine->{names}})],
                      [scalar(@{$theSubroutine->{code}}),
                       pack "V*", @{$theSubroutine->{code}}],
                      [scalar(@{$theSubroutine->{subrs}}),
                       pack "V*", @{$theSubroutine->{subrs}}],
                      [scalar(@{$theSubroutine->{locals}}),
                       packStringList(@{$theSubroutine->{locals}})],
                      [0,
                       $theSubroutine->{name}]);

    my $prop;
    my $body;
    foreach (@properties) {
        $prop .= pack("VV", $_->[0], length($_->[1]));
        $body .= $_->[1];
    }
    my $header = pack "V*", 1, $theSubroutine->{id}, length($prop) + length($body), scalar(@properties);
    $header.$prop.$body;
}

sub packStringList {
    my $result = "";
    foreach (@_) {
        $result .= chr(length($_)) . $_;
    }
    $result;
}

# Poor man's version of frexp. Split $arg into $mant and $exp so that
# $arg = $mant * 2**$exp and $arg in range [0.5, 1).
sub frexp {
    my $arg = shift;
    if ($arg == 0) {
        return (0,0);
    } else {
        my $exp = 0;
        my $mant = $arg;
        while ($exp > -1000 && $mant < 0.5) {
            $mant *= 2;
            --$exp;
        }
        while ($exp < +1000 && $mant >= 1.0) {
            $mant *= 0.5;
            ++$exp;
        }
        return ($mant, $exp);
    }
}

sub makeInstruction {
    my ($major, $minor, $argVal) = @_;
    16777216*$major + 65536*$minor + $argVal;
}

sub err {
    print STDERR "$inFN:$.: $_[0]\n";
    $numErrors++;
}
