/**
  *  \file ui/window.h
  *  \brief Top-Level Windows and Root Window Declarations
  *
  *  Top-Level windows used for the UI, and a root window to manage
  *  them.
  *
  *  (c) 2001-2010 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_UI_WINDOW_H
#define PCC_V2_UI_WINDOW_H

#include <queue>
#include "ui/ui.h"
#include "gfx/pixmap.h"
#include "gfx/gfxutil.h"
#include "gfx/multiclip.h"
#include "ui/skin.h"

class UIRoot : public UIBaseWidget {
    std::queue<int> pending_stops;
    GfxMultiClip multi_clip;       // regions that need redraw although not marked as such
    Ptr<GfxCanvas> screen;

    void performRegionUpdates();

 public:
    UIRoot(Ptr<GfxCanvas> screen);
    ~UIRoot();

    void performDeferredRedraws();

    virtual void          drawContent(GfxCanvas& canvas);                           // draw content of widget, end-users override this
    virtual void          drawEverything(GfxCanvas& canvas);                        // draw everything including children, bases override this
    virtual void          drawBelow(UIBaseWidget& child, GfxCanvas& canvas);        // draw everything below child, bases override this
    virtual GfxCanvas&    getCanvasFor(UIBaseWidget& child, ObjectHolder& mgr);     // get canvas to draw child on, bases override this

    virtual bool          handleEvent(const UIEvent& event, bool second_pass);       // can be overridden by whoever needs it, base provides default
    // virtual bool          handleCommand(const UICommand& cmd, UIBaseWidget& sender); // can be overridden by whoever needs it
    // virtual bool          isInside(GfxPoint pt);                                     // can be overridden by whoever needs it

    // virtual void          onStateChange(int bits, bool enabled); // ?

    virtual void          onAddRemoveChild(UIBaseWidget& the_widget, bool added);    // overridden in base

    virtual void          getLayoutInfo(LayoutInfo&);                                // overridden by everyone

    virtual void          onResize(GfxRect oldsize, bool by_layout);                 // overridden in base
    virtual void          onChildResize(UIBaseWidget& child, GfxRect oldsize);       // overridden in base

    virtual int           run();                 // overridden in root
    virtual void          stop(int arg);         // overridden in root

    // New utility functions:
    void                  add(UIBaseWidget& w);
    void                  close(UIBaseWidget& w);
    int                   runModal(UIBaseWidget& w);
    void                  centerWidget(UIBaseWidget& w);
    void                  moveWidgetToEdge(UIBaseWidget& w, int x_pos, int y_pos, int offset = 0);
    void                  markRedrawRegion(GfxRect r);
    void                  postTimerEvent();
};

class UIWindow;

class UIWindowSkin : public UISkin {
 public:
    UIWindowSkin(int bs);
};

class UIWindow : public UILayoutableWidget {
    static Ptr<GfxPixmap> frame_blue;
    friend class UIWindowSkin;

    UIWindowSkin skin;

    string_t name;
    int frame_style;
    int bg_style;
    int border_space;
 public:
    enum {
        BLUE_WINDOW = 0
    };

    UIWindow(string_t name, int fs, int bs, int space = 5);

    void drawContent(GfxCanvas& canvas);
    void adjustSize(GfxRect& size, bool to_inner);

    const string_t getName() const
        { return name; }

    static void drawWindow(GfxCanvas& canvas, const GfxRect& extent, int bg_style, string_t name, bool bg = true);
};

#endif
